@@ -1,4 +1,4 @@
-/* Copyright (C) 2007-2010 Open Information Security Foundation
+/* Copyright (C) 2007-2021 Open Information Security Foundation
  *
  * You can copy, redistribute or modify this Program under the terms of
  * the GNU General Public License version 2 as published by the Free
@@ -24,9 +24,7 @@
 #ifndef __APP_LAYER_SMTP_H__
 #define __APP_LAYER_SMTP_H__
 
-#include "decode-events.h"
 #include "util-decode-mime.h"
-#include "queue.h"
 #include "util-streaming-buffer.h"
 #include "rust.h"
 
@@ -43,6 +41,7 @@ enum {
     SMTP_DECODER_EVENT_NO_SERVER_WELCOME_MESSAGE,
     SMTP_DECODER_EVENT_TLS_REJECTED,
     SMTP_DECODER_EVENT_DATA_COMMAND_REJECTED,
+    SMTP_DECODER_EVENT_FAILED_PROTOCOL_CHANGE,
 
     /* MIME Events */
     SMTP_DECODER_EVENT_MIME_PARSE_FAILED,
@@ -84,55 +83,38 @@ typedef struct SMTPTransaction_ {
     /** the mime decoding parser state */
     MimeDecParseState *mime_state;
 
-    AppLayerDecoderEvents *decoder_events;          /**< per tx events */
-    DetectEngineState *de_state;
-
     /* MAIL FROM parameters */
     uint8_t *mail_from;
     uint16_t mail_from_len;
 
     TAILQ_HEAD(, SMTPString_) rcpt_to_list;  /**< rcpt to string list */
 
+    FileContainer files_ts;
+
     TAILQ_ENTRY(SMTPTransaction_) next;
 } SMTPTransaction;
 
 typedef struct SMTPConfig {
 
-    int decode_mime;
+    bool decode_mime;
     MimeDecConfig mime_config;
     uint32_t content_limit;
     uint32_t content_inspect_min_size;
     uint32_t content_inspect_window;
 
-    int raw_extraction;
+    bool raw_extraction;
 
     StreamingBufferConfig sbcfg;
 } SMTPConfig;
 
 typedef struct SMTPState_ {
+    AppLayerStateData state_data;
     SMTPTransaction *curr_tx;
     TAILQ_HEAD(, SMTPTransaction_) tx_list;  /**< transaction list */
     uint64_t tx_cnt;
     uint64_t toserver_data_count;
     uint64_t toserver_last_data_stamp;
 
-    /* current input that is being parsed */
-    const uint8_t *input;
-    int32_t input_len;
-    uint8_t direction;
-
-    /* original length of an input */
-    int32_t orig_input_len;
-
-    /* --parser details-- */
-    /** current line extracted by the parser from the call to SMTPGetline() */
-    const uint8_t *current_line;
-    /** length of the line in current_line.  Doesn't include the delimiter */
-    int32_t current_line_len;
-    uint8_t current_line_delimiter_len;
-    bool current_line_lf_found;
-    /* Consumed bytes till current line */
-    int32_t consumed;
     /* If rest of the bytes should be discarded in case of long line w/o LF */
     bool discard_till_lf_ts;
     bool discard_till_lf_tc;
@@ -164,7 +146,6 @@ typedef struct SMTPState_ {
 
     /* SMTP Mime decoding and file extraction */
     /** the list of files sent to the server */
-    FileContainer *files_ts;
     uint32_t file_track_id;
 } SMTPState;
 
