@@ -1,4 +1,4 @@
-/* Copyright (C) 2021 Open Information Security Foundation
+/* Copyright (C) 2021-2023 Open Information Security Foundation
  *
  * You can copy, redistribute or modify this Program under the terms of
  * the GNU General Public License version 2 as published by the Free
@@ -50,8 +50,8 @@
 
 extern int rule_reload;
 extern int engine_analysis;
-static int fp_engine_analysis_set = 0;
-int rule_engine_analysis_set = 0;
+static bool fp_engine_analysis_set = false;
+bool rule_engine_analysis_set = false;
 
 /**
  *  \brief Create the path if default-rule-path was specified
@@ -85,21 +85,9 @@ char *DetectLoadCompleteSigPath(const DetectEngineCtx *de_ctx, const char *sig_f
     /* Path not specified */
     if (PathIsRelative(sig_file)) {
         if (defaultpath) {
-            SCLogDebug("Default path: %s", defaultpath);
-            size_t path_len = sizeof(char) * (strlen(defaultpath) +
-                          strlen(sig_file) + 2);
-            path = SCMalloc(path_len);
+            path = PathMergeAlloc(defaultpath, sig_file);
             if (unlikely(path == NULL))
                 return NULL;
-            strlcpy(path, defaultpath, path_len);
-#if defined OS_WIN32 || defined __CYGWIN__
-            if (path[strlen(path) - 1] != '\\')
-                strlcat(path, "\\\\", path_len);
-#else
-            if (path[strlen(path) - 1] != '/')
-                strlcat(path, "/", path_len);
-#endif
-            strlcat(path, sig_file, path_len);
         } else {
             path = SCStrdup(sig_file);
             if (unlikely(path == NULL))
@@ -203,7 +191,7 @@ static int DetectLoadSigFile(DetectEngineCtx *de_ctx, char *sig_file,
                 }
             }
             if (rule_engine_analysis_set) {
-                EngineAnalysisRulesFailure(line, sig_file, lineno - multiline);
+                EngineAnalysisRulesFailure(de_ctx, line, sig_file, lineno - multiline);
             }
             if (!de_ctx->sigerror_ok) {
                 bad++;
@@ -301,8 +289,7 @@ int SigLoadSignatures(DetectEngineCtx *de_ctx, char *sig_file, int sig_file_excl
     }
 
     if (RunmodeGetCurrent() == RUNMODE_ENGINE_ANALYSIS) {
-        fp_engine_analysis_set = SetupFPAnalyzer();
-        rule_engine_analysis_set = SetupRuleAnalyzer();
+        SetupEngineAnalysis(de_ctx, &fp_engine_analysis_set, &rule_engine_analysis_set);
     }
 
     /* ok, let's load signature files from the general config */
@@ -387,12 +374,7 @@ int SigLoadSignatures(DetectEngineCtx *de_ctx, char *sig_file, int sig_file_excl
  end:
     gettimeofday(&de_ctx->last_reload, NULL);
     if (RunmodeGetCurrent() == RUNMODE_ENGINE_ANALYSIS) {
-        if (rule_engine_analysis_set) {
-            CleanupRuleAnalyzer();
-        }
-        if (fp_engine_analysis_set) {
-            CleanupFPAnalyzer();
-        }
+        CleanupEngineAnalysis(de_ctx);
     }
 
     DetectParseDupSigHashFree(de_ctx);
@@ -402,12 +384,12 @@ int SigLoadSignatures(DetectEngineCtx *de_ctx, char *sig_file, int sig_file_excl
 #define NLOADERS 4
 static DetectLoaderControl *loaders = NULL;
 static int cur_loader = 0;
-void TmThreadWakeupDetectLoaderThreads(void);
+static void TmThreadWakeupDetectLoaderThreads(void);
 static int num_loaders = NLOADERS;
 
 /** \param loader -1 for auto select
  *  \retval loader_id or negative in case of error */
-int DetectLoaderQueueTask(int loader_id, LoaderFunc Func, void *func_ctx)
+int DetectLoaderQueueTask(int loader_id, LoaderFunc Func, void *func_ctx, LoaderFreeFunc FreeFunc)
 {
     if (loader_id == -1) {
         loader_id = cur_loader;
@@ -427,6 +409,7 @@ int DetectLoaderQueueTask(int loader_id, LoaderFunc Func, void *func_ctx)
 
     t->Func = Func;
     t->ctx = func_ctx;
+    t->FreeFunc = FreeFunc;
 
     SCMutexLock(&loader->m);
     TAILQ_INSERT_TAIL(&loader->task_list, t, next);
@@ -444,14 +427,14 @@ int DetectLoadersSync(void)
 {
     SCLogDebug("waiting");
     int errors = 0;
-    int i;
-    for (i = 0; i < num_loaders; i++) {
-        int done = 0;
+    for (int i = 0; i < num_loaders; i++) {
+        bool done = false;
+
         DetectLoaderControl *loader = &loaders[i];
         while (!done) {
             SCMutexLock(&loader->m);
             if (TAILQ_EMPTY(&loader->task_list)) {
-                done = 1;
+                done = true;
             }
             SCMutexUnlock(&loader->m);
         }
@@ -461,7 +444,6 @@ int DetectLoadersSync(void)
             loader->result = 0;
         }
         SCMutexUnlock(&loader->m);
-
     }
     if (errors) {
         SCLogError("%d loaders reported errors", errors);
@@ -484,19 +466,17 @@ void DetectLoadersInit(void)
     (void)ConfGetInt("multi-detect.loaders", &setting);
 
     if (setting < 1 || setting > 1024) {
-        SCLogError("invalid multi-detect.loaders setting %" PRIdMAX, setting);
-        exit(EXIT_FAILURE);
+        FatalError("invalid multi-detect.loaders setting %" PRIdMAX, setting);
     }
-    num_loaders = (int32_t)setting;
 
+    num_loaders = (int32_t)setting;
     SCLogInfo("using %d detect loader threads", num_loaders);
 
     BUG_ON(loaders != NULL);
     loaders = SCCalloc(num_loaders, sizeof(DetectLoaderControl));
     BUG_ON(loaders == NULL);
 
-    int i;
-    for (i = 0; i < num_loaders; i++) {
+    for (int i = 0; i < num_loaders; i++) {
         DetectLoaderInit(&loaders[i]);
     }
 }
@@ -504,14 +484,11 @@ void DetectLoadersInit(void)
 /**
  * \brief Unpauses all threads present in tv_root
  */
-void TmThreadWakeupDetectLoaderThreads(void)
+static void TmThreadWakeupDetectLoaderThreads(void)
 {
-    ThreadVars *tv = NULL;
-    int i = 0;
-
     SCMutexLock(&tv_root_lock);
-    for (i = 0; i < TVT_MAX; i++) {
-        tv = tv_root[i];
+    for (int i = 0; i < TVT_MAX; i++) {
+        ThreadVars *tv = tv_root[i];
         while (tv != NULL) {
             if (strncmp(tv->name,"DL#",3) == 0) {
                 BUG_ON(tv->ctrl_cond == NULL);
@@ -521,8 +498,6 @@ void TmThreadWakeupDetectLoaderThreads(void)
         }
     }
     SCMutexUnlock(&tv_root_lock);
-
-    return;
 }
 
 /**
@@ -530,12 +505,9 @@ void TmThreadWakeupDetectLoaderThreads(void)
  */
 void TmThreadContinueDetectLoaderThreads(void)
 {
-    ThreadVars *tv = NULL;
-    int i = 0;
-
     SCMutexLock(&tv_root_lock);
-    for (i = 0; i < TVT_MAX; i++) {
-        tv = tv_root[i];
+    for (int i = 0; i < TVT_MAX; i++) {
+        ThreadVars *tv = tv_root[i];
         while (tv != NULL) {
             if (strncmp(tv->name,"DL#",3) == 0)
                 TmThreadContinue(tv);
@@ -544,11 +516,8 @@ void TmThreadContinueDetectLoaderThreads(void)
         }
     }
     SCMutexUnlock(&tv_root_lock);
-
-    return;
 }
 
-
 SC_ATOMIC_DECLARE(int, detect_loader_cnt);
 
 typedef struct DetectLoaderThreadData_ {
@@ -602,7 +571,7 @@ static TmEcode DetectLoader(ThreadVars *th_v, void *thread_data)
             int r = task->Func(task->ctx, ftd->instance);
             loader->result |= r;
             TAILQ_REMOVE(&loader->task_list, task, next);
-            SCFree(task->ctx);
+            task->FreeFunc(task->ctx);
             SCFree(task);
         }
 
@@ -630,27 +599,18 @@ static TmEcode DetectLoader(ThreadVars *th_v, void *thread_data)
 /** \brief spawn the detect loader manager thread */
 void DetectLoaderThreadSpawn(void)
 {
-    int i;
-    for (i = 0; i < num_loaders; i++) {
-        ThreadVars *tv_loader = NULL;
-
+    for (int i = 0; i < num_loaders; i++) {
         char name[TM_THREAD_NAME_MAX];
         snprintf(name, sizeof(name), "%s#%02d", thread_name_detect_loader, i+1);
 
-        tv_loader = TmThreadCreateCmdThreadByName(name,
-                "DetectLoader", 1);
-        BUG_ON(tv_loader == NULL);
-
+        ThreadVars *tv_loader = TmThreadCreateCmdThreadByName(name, "DetectLoader", 1);
         if (tv_loader == NULL) {
-            printf("ERROR: TmThreadsCreate failed\n");
-            exit(1);
+            FatalError("failed to create thread %s", name);
         }
         if (TmThreadSpawn(tv_loader) != TM_ECODE_OK) {
-            printf("ERROR: TmThreadSpawn failed\n");
-            exit(1);
+            FatalError("failed to create spawn %s", name);
         }
     }
-    return;
 }
 
 void TmModuleDetectLoaderRegister (void)
