@@ -28,13 +28,18 @@
 #include "conf.h"
 
 #include "decode.h"
+#include "packet.h"
 #include "flow.h"
 #include "stream-tcp.h"
 #include "app-layer.h"
 #include "app-layer-parser.h"
+#include "app-layer-frames.h"
 
 #include "detect.h"
+#include "detect-dsize.h"
 #include "detect-engine.h"
+#include "detect-engine-build.h"
+#include "detect-engine-frame.h"
 #include "detect-engine-profile.h"
 
 #include "detect-engine-alert.h"
@@ -58,6 +63,9 @@
 
 #include "util-validate.h"
 #include "util-detect.h"
+#include "util-profiling.h"
+
+#include "action-globals.h"
 
 typedef struct DetectRunScratchpad {
     const AppProto alproto;
@@ -83,6 +91,8 @@ static inline void DetectRulePacketRules(ThreadVars * const tv,
 static void DetectRunTx(ThreadVars *tv, DetectEngineCtx *de_ctx,
         DetectEngineThreadCtx *det_ctx, Packet *p,
         Flow *f, DetectRunScratchpad *scratch);
+static void DetectRunFrames(ThreadVars *tv, DetectEngineCtx *de_ctx, DetectEngineThreadCtx *det_ctx,
+        Packet *p, Flow *f, DetectRunScratchpad *scratch);
 static inline void DetectRunPostRules(ThreadVars *tv, DetectEngineCtx *de_ctx,
         DetectEngineThreadCtx *det_ctx, Packet * const p, Flow * const pflow,
         DetectRunScratchpad *scratch);
@@ -96,7 +106,10 @@ static void DetectRun(ThreadVars *th_v,
         Packet *p)
 {
     SCEnter();
-    SCLogDebug("pcap_cnt %"PRIu64, p->pcap_cnt);
+    SCLogDebug("p->pcap_cnt %" PRIu64 " direction %s pkt_src %s", p->pcap_cnt,
+            p->flow ? (FlowGetPacketDirection(p->flow, p) == TOSERVER ? "toserver" : "toclient")
+                    : "noflow",
+            PktSrcToString(p->pkt_src));
 
     /* bail early if packet should not be inspected */
     if (p->flags & PKT_NOPACKET_INSPECTION) {
@@ -132,6 +145,17 @@ static void DetectRun(ThreadVars *th_v,
 
     /* run tx/state inspection. Don't call for ICMP error msgs. */
     if (pflow && pflow->alstate && likely(pflow->proto == p->proto)) {
+        if (p->proto == IPPROTO_TCP) {
+            const TcpSession *ssn = p->flow->protoctx;
+            if (ssn && (ssn->flags & STREAMTCP_FLAG_APP_LAYER_DISABLED) == 0) {
+                // PACKET_PROFILING_DETECT_START(p, PROF_DETECT_TX);
+                DetectRunFrames(th_v, de_ctx, det_ctx, p, pflow, &scratch);
+                // PACKET_PROFILING_DETECT_END(p, PROF_DETECT_TX);
+            }
+        } else if (p->proto == IPPROTO_UDP) {
+            DetectRunFrames(th_v, de_ctx, det_ctx, p, pflow, &scratch);
+        }
+
         PACKET_PROFILING_DETECT_START(p, PROF_DETECT_TX);
         DetectRunTx(th_v, de_ctx, det_ctx, p, pflow, &scratch);
         PACKET_PROFILING_DETECT_END(p, PROF_DETECT_TX);
@@ -317,7 +341,7 @@ static inline void DetectPrefilterMergeSort(DetectEngineCtx *de_ctx,
                 goto final;
             }
             /* otherwise, if nonmpm is done jump to final for mpm
-             * mpm ptrs alrady updated */
+             * mpm ptrs already updated */
             if (unlikely(--n_cnt == 0)) {
                 final_ptr = mpm_ptr;
                 final_cnt = m_cnt;
@@ -352,15 +376,14 @@ static inline void DetectPrefilterMergeSort(DetectEngineCtx *de_ctx,
 /** \internal
  *  \brief build non-prefilter list based on the rule group list we've set.
  */
-static inline void
-DetectPrefilterBuildNonPrefilterList(DetectEngineThreadCtx *det_ctx,
-        const SignatureMask mask, const uint8_t alproto)
+static inline void DetectPrefilterBuildNonPrefilterList(
+        DetectEngineThreadCtx *det_ctx, const SignatureMask mask, const AppProto alproto)
 {
     for (uint32_t x = 0; x < det_ctx->non_pf_store_cnt; x++) {
         /* only if the mask matches this rule can possibly match,
          * so build the non_mpm array only for match candidates */
         const SignatureMask rule_mask = det_ctx->non_pf_store_ptr[x].mask;
-        const uint8_t rule_alproto = det_ctx->non_pf_store_ptr[x].alproto;
+        const AppProto rule_alproto = det_ctx->non_pf_store_ptr[x].alproto;
         if ((rule_mask & mask) == rule_mask &&
                 (rule_alproto == 0 || AppProtoEquals(rule_alproto, alproto))) {
             det_ctx->non_pf_id_array[det_ctx->non_pf_id_cnt++] = det_ctx->non_pf_store_ptr[x].id;
@@ -412,7 +435,6 @@ DetectPostInspectFileFlagsUpdate(Flow *f, const SigGroupHead *sgh, uint8_t direc
             flow_file_flags |= (FLOWFILE_NO_MAGIC_TS|FLOWFILE_NO_MAGIC_TC);
         }
 #endif
-#ifdef HAVE_NSS
         if (!(sgh->flags & SIG_GROUP_HEAD_HAVEFILEMD5)) {
             SCLogDebug("requesting disabling md5 for flow");
             flow_file_flags |= (FLOWFILE_NO_MD5_TS|FLOWFILE_NO_MD5_TC);
@@ -425,7 +447,6 @@ DetectPostInspectFileFlagsUpdate(Flow *f, const SigGroupHead *sgh, uint8_t direc
             SCLogDebug("requesting disabling sha256 for flow");
             flow_file_flags |= (FLOWFILE_NO_SHA256_TS|FLOWFILE_NO_SHA256_TC);
         }
-#endif
         if (!(sgh->flags & SIG_GROUP_HEAD_HAVEFILESIZE)) {
             SCLogDebug("requesting disabling filesize for flow");
             flow_file_flags |= (FLOWFILE_NO_SIZE_TS|FLOWFILE_NO_SIZE_TC);
@@ -730,7 +751,7 @@ static inline void DetectRulePacketRules(
     while (match_cnt--) {
         RULE_PROFILING_START(p);
         uint8_t alert_flags = 0;
-#ifdef PROFILING
+#ifdef PROFILE_RULES
         bool smatch = false; /* signature match */
 #endif
         s = next_s;
@@ -746,6 +767,9 @@ static inline void DetectRulePacketRules(
         if (s->app_inspect != NULL) {
             goto next; // handle sig in DetectRunTx
         }
+        if (s->frame_inspect != NULL) {
+            goto next; // handle sig in DetectRunFrame
+        }
 
         /* don't run mask check for stateful rules.
          * There we depend on prefilter */
@@ -754,13 +778,8 @@ static inline void DetectRulePacketRules(
             goto next;
         }
 
-        if (unlikely(sflags & SIG_FLAG_DSIZE)) {
-            if (likely(p->payload_len < s->dsize_low || p->payload_len > s->dsize_high)) {
-                SCLogDebug("kicked out as p->payload_len %u, dsize low %u, hi %u",
-                        p->payload_len, s->dsize_low, s->dsize_high);
-                goto next;
-            }
-        }
+        if (SigDsizePrefilter(p, s, sflags))
+            goto next;
 
         /* if the sig has alproto and the session as well they should match */
         if (likely(sflags & SIG_FLAG_APPLAYER)) {
@@ -778,7 +797,7 @@ static inline void DetectRulePacketRules(
             goto next;
         }
 
-#ifdef PROFILING
+#ifdef PROFILE_RULES
         smatch = true;
 #endif
         DetectRunPostMatch(tv, det_ctx, p, s);
@@ -810,7 +829,6 @@ static DetectRunScratchpad DetectRunSetup(
     p->alerts.discarded = 0;
     p->alerts.suppressed = 0;
 #endif
-    det_ctx->ticker++;
     det_ctx->filestore_cnt = 0;
     det_ctx->base64_decoded_len = 0;
     det_ctx->raw_stream_progress = 0;
@@ -904,7 +922,8 @@ static inline void DetectRunPostRules(
     /* see if we need to increment the inspect_id and reset the de_state */
     if (pflow && pflow->alstate) {
         PACKET_PROFILING_DETECT_START(p, PROF_DETECT_TX_UPDATE);
-        DeStateUpdateInspectTransactionId(pflow, scratch->flow_flags, (scratch->sgh == NULL));
+        AppLayerParserSetTransactionInspectId(pflow, pflow->alparser, pflow->alstate,
+                scratch->flow_flags, (scratch->sgh == NULL));
         PACKET_PROFILING_DETECT_END(p, PROF_DETECT_TX_UPDATE);
     }
 
@@ -949,7 +968,7 @@ void RuleMatchCandidateTxArrayInit(DetectEngineThreadCtx *det_ctx, uint32_t size
     DEBUG_VALIDATE_BUG_ON(det_ctx->tx_candidates);
     det_ctx->tx_candidates = SCCalloc(size, sizeof(RuleMatchCandidateTx));
     if (det_ctx->tx_candidates == NULL) {
-        FatalError(SC_ERR_MEM_ALLOC, "failed to allocate %"PRIu64" bytes",
+        FatalError("failed to allocate %" PRIu64 " bytes",
                 (uint64_t)(size * sizeof(RuleMatchCandidateTx)));
     }
     det_ctx->tx_candidates_size = size;
@@ -979,7 +998,7 @@ static int RuleMatchCandidateTxArrayExpand(DetectEngineThreadCtx *det_ctx, const
     uint32_t new_size = needed;
     void *ptmp = SCRealloc(det_ctx->tx_candidates, (new_size * sizeof(RuleMatchCandidateTx)));
     if (ptmp == NULL) {
-        FatalError(SC_ERR_MEM_ALLOC, "failed to expand to %"PRIu64" bytes",
+        FatalError("failed to expand to %" PRIu64 " bytes",
                 (uint64_t)(new_size * sizeof(RuleMatchCandidateTx)));
         // TODO can this be handled more gracefully?
     }
@@ -1030,7 +1049,7 @@ DetectRunTxSortHelper(const void *a, const void *b)
  *  detection.
  *
  *  \param stored_flags pointer to stored flags or NULL.
- *         If stored_flags is set it means we're continueing
+ *         If stored_flags is set it means we're continuing
  *         inspection from an earlier run.
  *
  *  \retval bool true sig matched, false didn't match
@@ -1052,7 +1071,7 @@ static bool DetectRunTxInspectRule(ThreadVars *tv,
     const int direction = (flow_flags & STREAM_TOSERVER) ? 0 : 1;
     uint32_t inspect_flags = stored_flags ? *stored_flags : 0;
     int total_matches = 0;
-    int file_no_match = 0;
+    uint16_t file_no_match = 0;
     bool retval = false;
     bool mpm_before_progress = false;   // is mpm engine before progress?
     bool mpm_in_progress = false;       // is mpm engine in a buffer we will revisit?
@@ -1115,20 +1134,15 @@ static bool DetectRunTxInspectRule(ThreadVars *tv,
             }
 
             /* run callback: but bypass stream callback if we can */
-            int match;
+            uint8_t match;
             if (unlikely(engine->stream && can->stream_stored)) {
                 match = can->stream_result;
                 TRACE_SID_TXS(s->id, tx, "stream skipped, stored result %d used instead", match);
             } else {
                 KEYWORD_PROFILING_SET_LIST(det_ctx, engine->sm_list);
-                if (engine->Callback) {
-                    match = engine->Callback(tv, de_ctx, det_ctx,
-                            s, engine->smd, f, flow_flags, alstate, tx->tx_ptr, tx->tx_id);
-                } else {
-                    BUG_ON(engine->v2.Callback == NULL);
-                    match = engine->v2.Callback(de_ctx, det_ctx, engine,
-                            s, f, flow_flags, alstate, tx->tx_ptr, tx->tx_id);
-                }
+                DEBUG_VALIDATE_BUG_ON(engine->v2.Callback == NULL);
+                match = engine->v2.Callback(
+                        de_ctx, det_ctx, engine, s, f, flow_flags, alstate, tx->tx_ptr, tx->tx_id);
                 TRACE_SID_TXS(s->id, tx, "engine %p match %d", engine, match);
                 if (engine->stream) {
                     can->stream_stored = true;
@@ -1219,6 +1233,11 @@ static bool DetectRunTxInspectRule(ThreadVars *tv,
     return retval;
 }
 
+#define NO_TX                                                                                      \
+    {                                                                                              \
+        NULL, 0, NULL, NULL, 0, 0, 0, 0, 0,                                                        \
+    }
+
 /** \internal
  *  \brief get a DetectTransaction object
  *  \retval struct filled with relevant info or all nulls/0s
@@ -1227,42 +1246,33 @@ static DetectTransaction GetDetectTx(const uint8_t ipproto, const AppProto alpro
         void *alstate, const uint64_t tx_id, void *tx_ptr, const int tx_end_state,
         const uint8_t flow_flags)
 {
-    uint64_t detect_flags;
     AppLayerTxData *txd = AppLayerParserGetTxData(ipproto, alproto, tx_ptr);
-    if (likely(txd != NULL)) {
-        detect_flags = (flow_flags & STREAM_TOSERVER) ? txd->detect_flags_ts : txd->detect_flags_tc;
-    } else {
-        detect_flags = 0;
+    if (unlikely(txd == NULL)) {
+        DetectTransaction no_tx = NO_TX;
+        return no_tx;
     }
+    uint64_t detect_flags =
+            (flow_flags & STREAM_TOSERVER) ? txd->detect_flags_ts : txd->detect_flags_tc;
     if (detect_flags & APP_LAYER_TX_INSPECTED_FLAG) {
         SCLogDebug("%"PRIu64" tx already fully inspected for %s. Flags %016"PRIx64,
                 tx_id, flow_flags & STREAM_TOSERVER ? "toserver" : "toclient",
                 detect_flags);
-        DetectTransaction no_tx = { NULL, 0, NULL, NULL, 0, 0, 0, 0, 0, };
+        DetectTransaction no_tx = NO_TX;
         return no_tx;
     }
     if (detect_flags & APP_LAYER_TX_SKIP_INSPECT_FLAG) {
         SCLogDebug("%" PRIu64 " tx should not be inspected in direction %s. Flags %016" PRIx64,
                 tx_id, flow_flags & STREAM_TOSERVER ? "toserver" : "toclient", detect_flags);
-        DetectTransaction no_tx = {
-            NULL,
-            0,
-            NULL,
-            NULL,
-            0,
-            0,
-            0,
-            0,
-            0,
-        };
+        DetectTransaction no_tx = NO_TX;
         return no_tx;
     }
 
     const int tx_progress = AppLayerParserGetStateProgress(ipproto, alproto, tx_ptr, flow_flags);
     const int dir_int = (flow_flags & STREAM_TOSERVER) ? 0 : 1;
-    DetectEngineState *tx_de_state = AppLayerParserGetTxDetectState(ipproto, alproto, tx_ptr);
+    DetectEngineState *tx_de_state = txd->de_state;
     DetectEngineStateDirection *tx_dir_state = tx_de_state ? &tx_de_state->dir_state[dir_int] : NULL;
     uint64_t prefilter_flags = detect_flags & APP_LAYER_TX_PREFILTER_MASK;
+    DEBUG_VALIDATE_BUG_ON(prefilter_flags & APP_LAYER_TX_RESERVED_FLAGS);
 
     DetectTransaction tx = {
                             .tx_ptr = tx_ptr,
@@ -1430,6 +1440,10 @@ static void DetectRunTx(ThreadVars *tv,
                     DetectRunTxSortHelper);
         }
 
+#ifdef PROFILING
+        if (array_idx >= de_ctx->profile_match_logging_threshold)
+            RulesDumpTxMatchArray(det_ctx, scratch->sgh, p, tx.tx_id, array_idx, x);
+#endif
         det_ctx->tx_id = tx.tx_id;
         det_ctx->tx_id_set = true;
         det_ctx->p = p;
@@ -1472,7 +1486,7 @@ static void DetectRunTx(ThreadVars *tv,
 
             if (inspect_flags) {
                 /* continue previous inspection */
-                SCLogDebug("%p/%"PRIu64" Continueing sid %u", tx.tx_ptr, tx.tx_id, s->id);
+                SCLogDebug("%p/%" PRIu64 " Continuing sid %u", tx.tx_ptr, tx.tx_id, s->id);
             } else {
                 /* start new inspection */
                 SCLogDebug("%p/%"PRIu64" Start sid %u", tx.tx_ptr, tx.tx_id, s->id);
@@ -1511,6 +1525,7 @@ static void DetectRunTx(ThreadVars *tv,
         }
         if (tx.prefilter_flags != tx.prefilter_flags_orig) {
             new_detect_flags |= tx.prefilter_flags;
+            DEBUG_VALIDATE_BUG_ON(new_detect_flags & APP_LAYER_TX_RESERVED_FLAGS);
             SCLogDebug("%p/%"PRIu64" updated prefilter flags %016"PRIx64" "
                     "(was: %016"PRIx64") for direction %s. Flag %016"PRIx64,
                     tx.tx_ptr, tx.tx_id, tx.prefilter_flags, tx.prefilter_flags_orig,
@@ -1521,6 +1536,7 @@ static void DetectRunTx(ThreadVars *tv,
                 (new_detect_flags | tx.detect_flags) != tx.detect_flags)
         {
             new_detect_flags |= tx.detect_flags;
+            DEBUG_VALIDATE_BUG_ON(new_detect_flags & APP_LAYER_TX_RESERVED_FLAGS);
             SCLogDebug("%p/%"PRIu64" Storing new flags %016"PRIx64" (was %016"PRIx64")",
                     tx.tx_ptr, tx.tx_id, new_detect_flags, tx.detect_flags);
 
@@ -1534,10 +1550,138 @@ next:
     }
 }
 
+static void DetectRunFrames(ThreadVars *tv, DetectEngineCtx *de_ctx, DetectEngineThreadCtx *det_ctx,
+        Packet *p, Flow *f, DetectRunScratchpad *scratch)
+{
+    const SigGroupHead *const sgh = scratch->sgh;
+    const AppProto alproto = f->alproto;
+
+    FramesContainer *frames_container = AppLayerFramesGetContainer(f);
+    if (frames_container == NULL) {
+        return;
+    }
+    Frames *frames;
+    if (PKT_IS_TOSERVER(p)) {
+        frames = &frames_container->toserver;
+    } else {
+        frames = &frames_container->toclient;
+    }
+
+    for (uint32_t idx = 0; idx < frames->cnt; idx++) {
+        SCLogDebug("frame %u", idx);
+        Frame *frame = FrameGetByIndex(frames, idx);
+        if (frame == NULL) {
+            continue;
+        }
+
+        det_ctx->frame_inspect_progress = 0;
+        uint32_t array_idx = 0;
+        uint32_t total_rules = det_ctx->match_array_cnt;
+
+        /* run prefilter engines and merge results into a candidates array */
+        if (sgh->frame_engines) {
+            //            PACKET_PROFILING_DETECT_START(p, PROF_DETECT_PF_TX);
+            DetectRunPrefilterFrame(det_ctx, sgh, p, frames, frame, alproto);
+            //            PACKET_PROFILING_DETECT_END(p, PROF_DETECT_PF_TX);
+            SCLogDebug("%p/%" PRIi64 " rules added from prefilter: %u candidates", frame, frame->id,
+                    det_ctx->pmq.rule_id_array_cnt);
+
+            total_rules += det_ctx->pmq.rule_id_array_cnt;
+
+            if (!(RuleMatchCandidateTxArrayHasSpace(
+                        det_ctx, total_rules))) { // TODO is it safe to overload?
+                RuleMatchCandidateTxArrayExpand(det_ctx, total_rules);
+            }
+
+            for (uint32_t i = 0; i < det_ctx->pmq.rule_id_array_cnt; i++) {
+                const Signature *s = de_ctx->sig_array[det_ctx->pmq.rule_id_array[i]];
+                const SigIntId id = s->num;
+                det_ctx->tx_candidates[array_idx].s = s;
+                det_ctx->tx_candidates[array_idx].id = id;
+                det_ctx->tx_candidates[array_idx].flags = NULL;
+                det_ctx->tx_candidates[array_idx].stream_reset = 0;
+                array_idx++;
+            }
+            PMQ_RESET(&det_ctx->pmq);
+        }
+        /* merge 'state' rules from the regular prefilter */
+        uint32_t x = array_idx;
+        for (uint32_t i = 0; i < det_ctx->match_array_cnt; i++) {
+            const Signature *s = det_ctx->match_array[i];
+            if (s->frame_inspect != NULL) {
+                const SigIntId id = s->num;
+                det_ctx->tx_candidates[array_idx].s = s;
+                det_ctx->tx_candidates[array_idx].id = id;
+                det_ctx->tx_candidates[array_idx].flags = NULL;
+                det_ctx->tx_candidates[array_idx].stream_reset = 0;
+                array_idx++;
+
+                SCLogDebug("%p/%" PRIi64 " rule %u (%u) added from 'match' list", frame, frame->id,
+                        s->id, id);
+            }
+        }
+        SCLogDebug("%p/%" PRIi64 " rules added from 'match' list: %u", frame, frame->id,
+                array_idx - x);
+        (void)x;
+
+        /* run rules: inspect the match candidates */
+        for (uint32_t i = 0; i < array_idx; i++) {
+            const Signature *s = det_ctx->tx_candidates[i].s;
+
+            /* deduplicate: rules_array is sorted, but not deduplicated.
+             * As they are back to back in that case we can check for it
+             * here. We select the stored state one as that comes first
+             * in the array. */
+            while ((i + 1) < array_idx &&
+                    det_ctx->tx_candidates[i].s == det_ctx->tx_candidates[i + 1].s) {
+                i++;
+            }
+            SCLogDebug("%p/%" PRIi64 " inspecting: sid %u (%u)", frame, frame->id, s->id, s->num);
+
+            /* start new inspection */
+            SCLogDebug("%p/%" PRIi64 " Start sid %u", frame, frame->id, s->id);
+
+            /* call individual rule inspection */
+            RULE_PROFILING_START(p);
+            int r = DetectRunInspectRuleHeader(p, f, s, s->flags, s->proto.flags);
+            if (r == 1) {
+                r = DetectRunFrameInspectRule(tv, det_ctx, s, f, p, frames, frame);
+                if (r == 1) {
+                    /* match */
+                    DetectRunPostMatch(tv, det_ctx, p, s);
+
+                    const uint8_t alert_flags =
+                            (PACKET_ALERT_FLAG_STATE_MATCH | PACKET_ALERT_FLAG_FRAME);
+                    det_ctx->flags |= DETECT_ENGINE_THREAD_CTX_FRAME_ID_SET;
+                    det_ctx->frame_id = frame->id;
+                    SCLogDebug(
+                            "%p/%" PRIi64 " sig %u (%u) matched", frame, frame->id, s->id, s->num);
+                    AlertQueueAppend(det_ctx, s, p, frame->tx_id, alert_flags);
+                }
+            }
+            DetectVarProcessList(det_ctx, p->flow, p);
+            RULE_PROFILING_END(det_ctx, s, r, p);
+        }
+
+        /* update Frame::inspect_progress here instead of in the code above. The reason is that a
+         * frame might be used more than once in buffers with transforms. */
+        if (frame->inspect_progress < det_ctx->frame_inspect_progress) {
+            frame->inspect_progress = det_ctx->frame_inspect_progress;
+            SCLogDebug("frame->inspect_progress: %" PRIu64 " -> updated", frame->inspect_progress);
+        } else {
+            SCLogDebug(
+                    "frame->inspect_progress: %" PRIu64 " -> not updated", frame->inspect_progress);
+        }
+
+        SCLogDebug("%p/%" PRIi64 " rules inspected, running cleanup", frame, frame->id);
+        InspectionBufferClean(det_ctx);
+    }
+}
+
 static DetectEngineThreadCtx *GetTenantById(HashTable *h, uint32_t id)
 {
     /* technically we need to pass a DetectEngineThreadCtx struct with the
-     * tentant_id member. But as that member is the first in the struct, we
+     * tenant_id member. But as that member is the first in the struct, we
      * can use the id directly. */
     return HashTableLookup(h, &id, 0);
 }
@@ -1546,20 +1690,19 @@ static void DetectFlow(ThreadVars *tv,
                        DetectEngineCtx *de_ctx, DetectEngineThreadCtx *det_ctx,
                        Packet *p)
 {
+    Flow *const f = p->flow;
+
     if (p->flags & PKT_NOPACKET_INSPECTION) {
         /* hack: if we are in pass the entire flow mode, we need to still
          * update the inspect_id forward. So test for the condition here,
          * and call the update code if necessary. */
-        const int pass = ((p->flow->flags & FLOW_NOPACKET_INSPECTION));
+        const int pass = ((f->flags & FLOW_NOPACKET_INSPECTION));
         if (pass) {
-            uint8_t flags;
-            if (p->flowflags & FLOW_PKT_TOSERVER) {
-                flags = STREAM_TOSERVER;
-            } else {
-                flags = STREAM_TOCLIENT;
+            uint8_t flags = STREAM_FLAGS_FOR_PACKET(p);
+            flags = FlowGetDisruptionFlags(f, flags);
+            if (f->alstate) {
+                AppLayerParserSetTransactionInspectId(f, f->alparser, f->alstate, flags, true);
             }
-            flags = FlowGetDisruptionFlags(p->flow, flags);
-            DeStateUpdateInspectTransactionId(p->flow, flags, true);
         }
         SCLogDebug("p->pcap %"PRIu64": no detection on packet, "
                 "PKT_NOPACKET_INSPECTION is set", p->pcap_cnt);
@@ -1569,8 +1712,8 @@ static void DetectFlow(ThreadVars *tv,
     /* we check the flow drop here, and not the packet drop. This is
      * to allow stream engine "invalid" drop packets to still be
      * evaluated by the stream event rules. */
-    if (p->flow->flags & FLOW_ACTION_DROP) {
-        DEBUG_VALIDATE_BUG_ON(!(PKT_IS_PSEUDOPKT(p)) && !PACKET_TEST_ACTION(p, ACTION_DROP));
+    if (f->flags & FLOW_ACTION_DROP) {
+        DEBUG_VALIDATE_BUG_ON(!(PKT_IS_PSEUDOPKT(p)) && !PacketCheckAction(p, ACTION_DROP));
         SCReturn;
     }
 
@@ -1583,10 +1726,8 @@ static void DetectNoFlow(ThreadVars *tv,
                          DetectEngineCtx *de_ctx, DetectEngineThreadCtx *det_ctx,
                          Packet *p)
 {
-    /* No need to perform any detection on this packet, if the the given flag is set.*/
-    if ((p->flags & PKT_NOPACKET_INSPECTION) ||
-        (PACKET_TEST_ACTION(p, ACTION_DROP)))
-    {
+    /* No need to perform any detection on this packet, if the given flag is set.*/
+    if ((p->flags & PKT_NOPACKET_INSPECTION) || (PacketCheckAction(p, ACTION_DROP))) {
         return;
     }
 
@@ -1653,6 +1794,15 @@ TmEcode Detect(ThreadVars *tv, Packet *p, void *data)
     } else {
         DetectNoFlow(tv, de_ctx, det_ctx, p);
     }
+
+#ifdef PROFILE_RULES
+    /* aggregate statistics */
+    if (SCTIME_SECS(p->ts) != det_ctx->rule_perf_last_sync) {
+        SCProfilingRuleThreatAggregate(det_ctx);
+        det_ctx->rule_perf_last_sync = SCTIME_SECS(p->ts);
+    }
+#endif
+
     return TM_ECODE_OK;
 error:
     return TM_ECODE_FAILED;
